# -*- coding: utf-8 -*-


import os
import sys
import traceback
from OpenSSL import crypto
import uuid
from datetime import datetime
from flask import (Flask, session, request, redirect, url_for, render_template,
                    jsonify, session, abort, send_file, flash, g)
from slapos.certificate_authority.web.start_web import app, db
from slapos.certificate_authority.web.tools import (find_user, find_or_create_user,
                    get_string_num, CertificateTools)
from flask_user import login_required, current_user
from flask_login import logout_user  #, login_user, current_user, login_required
from slapos.certificate_authority.web.models import (UserProfileForm, Certificate)
from slapos.runner.utils import tail

cert_tools = CertificateTools()

@app.errorhandler(403)
def error403(res):
  return "403: Forbidden. Your are not allowed to access %s" % res, 403
  return response

@app.errorhandler(404)
def error404(msg=""):
  return "404: Resource not found.\n%s\n" % msg, 404

@app.errorhandler(501)
def error501(msg=""):
  return "501: Internal Error. %s\n" % msg, 501

@app.errorhandler(412)
def error412(msg):
  return "412: Precondition Failed: %s.\n" % msg, 412

@app.errorhandler(400)
def error400(msg):
  return "400: Bad Request", 400

def writefile(path, content, mode=0640):
  with open(path, 'w') as fd:
    fd.write(content)
  os.chmod(path, mode)


@app.before_request
def before_request():
  if request.path.startswith('/static/') \
    or request.path.startswith == '/get/':
    return
  if request.path == '/configure' or request.path == '/setpassword':
    # check if password file exists, if yes go to index
    if find_user('admin'):
      return redirect(url_for('home'))
    return
  # XXX - using hard username
  if not find_user('admin'):
    return redirect(url_for('configure'))
  g.user = current_user

@app.route('/')
def index():
  # page to list certificates, also connection link
  data_list = cert_tools.getCertificateList()
  return render_template("index.html", data_list=data_list)


@app.route('/get/<string:name>', methods=['GET'])
def getfile(name):
  ca_name = os.path.basename(app.config.ca.certificate)
  crl_name = os.path.basename(app.config.ca.ca_crl)
  if name == ca_name:
    return send_file(app.config.ca.certificate,
                      attachment_filename=ca_name,
                      as_attachment=True)
  elif name == crl_name:
    return send_file(app.config.ca.ca_crl,
                      attachment_filename=crl_name,
                      as_attachment=True)
  else:
    # name is like f9c4-4a04-8bad-70bb45c6406e.cert.pem
    cert_filename = cert_tools.getFilenameFromHash(name)
    cert_file = os.path.join(app.config.cert_dir, cert_filename)
    if os.path.exists(cert_file) and os.path.isfile(cert_file):
      filename = os.path.basename(cert_file)
      return send_file(cert_file,
                        attachment_filename=filename,
                        as_attachment=True)

  return abort(404, name)

@app.route('/logout')
def logout():
    logout_user()
    return redirect(url_for('home'))

@app.route('/configure', methods=['GET'])
def configure():
  return render_template("configure.html")

@app.route('/setpassword', methods=['POST'])
def setpassword():
  username = 'admin'
  password = request.form.get('password', '').encode('utf-8')
  if not password:
    return abort(412, "password not set or empty")

  find_or_create_user(
    "Admin",
    "admin",
    "admin@example.com",
    username,
    password)
  logout_user()
  return redirect(url_for('home'))






# Routes for certificate Authority
def signcert(key, redirect_to=''):
  req_file = os.path.join(app.config.req_dir, '%s.csr.pem' % key)
  if os.path.exists(req_file) and os.path.isfile(req_file):
    cert_id = key.split('+')[0]
    try:
      cert_tools.signCertificate(cert_id, req_file)
    except Exception, e:
      # XXX - need to check this
      raise
      # return abort(501, str(e))
  else:
    return abort(403)
  flash('Certificate is signed!', 'success')
  if redirect_to:
    return redirect(url_for(redirect_to))
  return "Certificate is signed"

@app.route('/request', methods=['POST'])
def do_request():
  csr_content = request.form.get('csr', '').encode('utf-8')
  if not csr_content:
    return abort(400)

  try:
    req = app.config.ca.readCertificateRequest(csr_content)
    # XXX - Maybe more check if the csr is valid with CA
  except crypto.Error, e:
    return abort(412, str(e))

  cert_id = str(uuid.uuid1())
  csr_keyfile = '%s+%s.csr.pem' % (cert_id, str(uuid.uuid4()))
  request_file = os.path.join(app.config.req_dir, csr_keyfile)

  if os.path.exists(request_file):
    # The request already exist
    raise Exception("Certificate Signature Request file should be unique")
  try:
    writefile(request_file, csr_content)
  except OSError, e:
    raise

  return cert_id

@app.route('/signcert', methods=['POST'])
def do_signcert():
  """
    This method should be called by a list of host allowed, can be used to sign with command line
    For security, it's good to only allow local ip (127.0.0.1) or ::1
  """
  key = request.form.get('key', '').encode('utf-8')
  if not key:
    return abort(400)
  remote_client = request.remote_addr
  x_forwarded_for_list = request.headers.getlist("X-Forwarded-For")
  if remote_client not in app.config.trusted_host_list or \
      (x_forwarded_for_list and x_forwarded_for_list[0] not in app.config.trusted_host_list):

    return abort(403)  # Forbidden

  return signcert(key)

@app.route('/renewcert', methods=['POST'])
def renewcert(serial):
  """
  this method is used to renew expired certificate.
  should recieve csr and old cert serial.
  """
  csr_content = request.form.get('csr', '').encode('utf-8')
  serial = request.form.get('serial', '').encode('utf-8')
  return abort(200, "Done. The method is not implemented yet!")

@app.route('/requestrevoke', methods=['POST'])
def do_revoke():
  """
    Revoke method required certificat: serial(Hex), name and reason
    'name' is used here to add more verification
    'reason' is a message that say why this certificate should be revoked
  """
  serial = request.form.get('serial', '').upper()
  name = request.form.get('name', '')
  # A msg explaining why the certificate should be revoked
  reason = request.form.get('reason', '')
  if not serial or not name or not reason:
    # Bad parameters
    return abort(400)

  if len(serial) == 1:
    serial = '0%s' % serial

  if cert_tools.addRevokeRequest(serial, name, reason):
    return "Certificate revocation created"
  else:
    return abort(412, "Parameters not valid for the request")







#Manage routes (access admin)

@app.route('/manage', methods=['GET'])
@login_required
def manage_home():
  data_list = cert_tools.getCertificateRequestList()
  c = dict(
    cert=get_string_num(cert_tools.countCert()),
    csr=get_string_num(cert_tools.countCsr()),
    crl=get_string_num(cert_tools.countRevokedCert()),
    crr=get_string_num(cert_tools.countRevocationRequest())
  )
  return render_template('manage_page.html', data_list=data_list, counter=c)

@app.route('/profile', methods=['GET', 'POST'])
@login_required
def user_profile():
    form = UserProfileForm(request.form, obj=current_user)
    if request.method == 'POST' and form.validate():
        # Copy form fields to user_profile fields
        del form.username  # revove username from recieved form
        form.populate_obj(current_user)
        db.session.commit()

        # Redirect to home page
        return redirect(url_for('home'))

    return render_template('user_profile.html',
                           form=form)

@app.route('/viewcert', methods=['GET'])
@login_required
def viewcert():
  type_dict = {
    'cacert': os.path.dirname(app.config.ca.certificate),
    'X509': app.config.cert_dir,
    'req': app.config.req_dir,
    'crl': app.config.crl_dir,
  }
  filename = request.args.get('name')
  request_type = request.args.get('type')
  content = cert_type = ""
  name = "Requested file not found!"
  if filename and request_type in type_dict.keys():
    path = os.path.join(type_dict[request_type], filename)
    if os.path.exists(path) and os.path.isfile(path):
      name = filename
      if request_type in ['X509', 'cacert']:
        cert_type = 'Certificate'
        content = app.config.ca.getX509Text(app.config.ca.openssl_binary, path)
      elif request_type == 'req':
        cert_type = 'CSR'
        content = app.config.ca.getCsrText(app.config.ca.openssl_binary, path)
      elif request_type == 'crl':
        cert_type = 'CRL'
        content = app.config.ca.getCrlText(app.config.ca.openssl_binary, path)

  return render_template('view_cert.html', content=content, name=name, cert_type=cert_type)

@app.route('/signcert_web', methods=['GET'])
@login_required
def do_signcert_web():
  filename = request.args.get('name', '').encode('utf-8')
  if not filename or len(filename) < 9:
    return abort(412)
  return signcert(filename[:-8], 'manage')

@app.route('/deletecsr', methods=['GET'])
@login_required
def deletecsr():
  """
    Delete certificate signature request file
  """
  filename = request.args.get('name', '').encode('utf-8')
  if not filename:
    return abort(412)
  req_file = os.path.join(app.config.req_dir, filename)
  if os.path.exists(req_file):
    os.unlink(req_file)
    flash('CSR file "%s" removed!' % filename, 'success')
    return redirect(url_for('manage'))
  else:
    return abort(404)

@app.route('/signed_certs', methods=['GET'])
@login_required
def signed_certificate_list():
  # page to list certificates, also connection link
  data_list = cert_tools.getCertificateList(False)
  return render_template("signed_certs.html", data_list=data_list)

@app.route('/revoked_certs', methods=['GET'])
@login_required
def revoked_certificate_list():
  # page to list certificates, also connection link
  data_list = cert_tools.getRevokedCertificateList()
  return render_template("revoked_certs.html", data_list=data_list)

@app.route('/revoke_cert', methods=['GET', 'POST'])
@login_required
def revoke_cert():
  redirect_url = 'signed_certs'
  serial = request.args.get('serial', '')
  if not serial:
    return abort(412, "serial is invalid")
  request_id = request.args.get('request_id', None)
  if request_id:
    redirect_url = 'revocation_requests'

  if cert_tools.revokeCertificate(serial, request_id):
    flash('Certificate was revoked!', 'success')
    return redirect(url_for(redirect_url))

  return abort(404)

@app.route('/revocation_requests', methods=['GET'])
@login_required
def revocation_request_list():
  data_list = cert_tools.getRevocationRequestList()
  return render_template("revocation_requests.html", data_list=data_list)

@app.route('/reject_revocation_request', methods=['GET'])
@login_required
def reject_revocation_request():
  req_id = request.args.get('request_id', '')
  if not req_id:
    return abort(412)
  if cert_tools.cancelRevocationRequest(req_id):
    flash('Revocation Request Cancelled!', 'success')
  else:
    flash('Revocation Request not found!', 'info')
  return redirect(url_for('revocation_requests'))


@app.route('/view_logs', methods=['GET'])
@login_required
def view_logs():
  content = ""
  full_log = request.args.get('full', '')
  if os.path.exists(app.config.log_file):
    with open(app.config.log_file) as f:
      if full_log == 'true':
        content = f.read()
      else:
        content = tail(f, 500)
  return render_template("view_logs.html", content=content, filename=app.config.log_file)

@app.route('/generate_crl', methods=['GET'])
@login_required
def generate_crl():
  app.config.ca.genCertificateRevocationList()
  return redirect(url_for('home'))


app.add_url_rule('/', 'home', index)
app.add_url_rule('/manage', 'manage', manage_home)
app.add_url_rule('/signed_certs', 'signed_certs', signed_certificate_list)
app.add_url_rule('/revocation_requests', 'revocation_requests', revocation_request_list)